import logging
from typing import Callable
from unittest.mock import MagicMock

import psga


def test_action_implicit_name():
    @psga.action()
    def my_action(_):
        pass

    assert isinstance(my_action, Callable)
    assert hasattr(my_action, "name")
    assert getattr(my_action, "name") == my_action.name

    assert hasattr(my_action, "keys")
    assert getattr(my_action, "keys") == None


def test_action_explicit_name():
    @psga.action(name="my_action")
    def my_action(_):
        pass

    assert isinstance(my_action, Callable)
    assert hasattr(my_action, "name")
    assert getattr(my_action, "name") == "my_action"
    assert getattr(my_action, "name") == my_action.name

    assert hasattr(my_action, "keys")
    assert getattr(my_action, "keys") == None


def test_action_explicit_keys():
    KEYS = ["my_key"]

    @psga.action(keys=KEYS)
    def my_action(_):
        pass

    assert isinstance(my_action, Callable)
    assert hasattr(my_action, "name")
    assert getattr(my_action, "name") == my_action.name

    assert hasattr(my_action, "keys")
    assert getattr(my_action, "keys") == KEYS


def test_action_explicit_name_explicit_keys():
    KEYS = ["my_key"]

    @psga.action(name="my_action", keys=KEYS)
    def my_action(_):
        pass

    assert isinstance(my_action, Callable)
    assert hasattr(my_action, "name")
    assert getattr(my_action, "name") == "my_action"
    assert getattr(my_action, "name") == my_action.name

    assert hasattr(my_action, "keys")
    assert getattr(my_action, "keys") == KEYS


def test_dispatcher_register_one():
    handler1_invoked = 0

    @psga.action(name="event")
    def handler1(values):
        nonlocal handler1_invoked
        handler1_invoked += 1
        assert values == {1: 1}

    dispatcher = psga.Dispatcher()
    dispatcher.register(handler1)

    assert not dispatcher.dispatch("wrong_event", {1: 1})
    assert handler1_invoked == 0

    assert dispatcher.dispatch("event", {1: 1})
    assert handler1_invoked == 1

    assert dispatcher.dispatch(handler1.name, {1: 1})
    assert handler1_invoked == 2


def test_dispatcher_register_multiple():
    handler1_invoked = handler2_invoked = 0

    @psga.action()
    def handler1(values):
        nonlocal handler1_invoked
        handler1_invoked += 1
        assert values == {1: 1}

    @psga.action(handler1.name)
    def handler2(values):
        nonlocal handler2_invoked
        handler2_invoked += 1
        assert values == {1: 1}

    dispatcher = psga.Dispatcher()
    dispatcher.register(handler1).register(handler2)

    assert not dispatcher.dispatch("wrong_event", {1: 1})
    assert handler1_invoked == 0
    assert handler2_invoked == 0

    assert dispatcher.dispatch(handler1.name, {1: 1})
    assert handler1_invoked == 1
    assert handler2_invoked == 1

    assert dispatcher.dispatch(handler2.name, {1: 1})
    assert handler1_invoked == 2
    assert handler2_invoked == 2


def test_controller():
    class _MyController(psga.Controller):
        answer = 0

        @psga.action(name="universal_question")
        def on_ask(self, values):
            _MyController.answer = values

    dispatcher = psga.Dispatcher()
    controller = _MyController(dispatcher)

    assert controller.answer == 0
    dispatcher.dispatch(controller.on_ask.name, 42)
    assert controller.answer == 42


def test_dispatcher_loop():
    logging.getLogger("PSGA").setLevel(logging.DEBUG)

    mock_window = MagicMock()
    mock_window.configure_mock(
        **{
            "read.side_effect": [
                ("Ok", {1: 1}),
                ("Unknown", {2: 2}),
                ("Unregistered", {3: 3}),
                (("-TABLE KEY-", "+CLICKED+", (3, 3)), {4: 4}),
                ("Unknown menu::menu_key", {5: 5}),
                ("-INPUT 1-", {6: 6}),
                ("-BUTTON SAVE-", {6: 6}),
                ("Exit", {6: 6}),
            ]
        }
    )

    handler1_invoked = 0
    handler2_invoked = 0
    handler3_invoked = 0
    handler4_invoked = 0
    handler5_invoked = 0

    @psga.action(name="Ok")
    def handler1(values):
        nonlocal handler1_invoked
        handler1_invoked += 1
        assert values == {1: 1}

    @psga.action(name="Unregistered")
    def handler4(values):
        nonlocal handler4_invoked
        handler4_invoked += 1
        assert values == {3: 3}

    @psga.action(name="-TABLE KEY-")
    def handler2(values):
        nonlocal handler2_invoked
        handler2_invoked += 1
        assert values == {4: 4}

    @psga.action(name="menu_key")
    def handler3(values):
        nonlocal handler3_invoked
        handler3_invoked += 1
        assert values == {5: 5}

    @psga.action(keys=["-INPUT 1-", "-BUTTON SAVE-"])
    def handler5(values):
        nonlocal handler5_invoked
        handler5_invoked += 1
        assert values == {6: 6}

    dispatcher = psga.Dispatcher()
    dispatcher.register(handler1)
    dispatcher.register(handler2)
    dispatcher.register(handler3)
    dispatcher.register(handler5)

    dispatcher.loop(mock_window)

    assert handler1_invoked == 1
    assert handler2_invoked == 1
    assert handler3_invoked == 1
    assert handler4_invoked == 0
    assert handler5_invoked == 2



# Copyright 2023 Francis Meyvis <psga@mikmak.fun>

"""Minimalistic Controller (as in the MVC paradigm) for PySimpleGUI."""

import functools
import logging
from typing import Callable, Dict, Hashable, List, Optional, Protocol

import PySimpleGUI as sg
from typing_extensions import Self


class Action(Protocol):
    """Bundles the event's name and handler"""

    # pylint: disable=too-few-public-methods

    name: str
    keys: Optional[List[Hashable]]

    def __call__(self, values=None):
        """"""


def action(name: Optional[str] = None, keys: Optional[List[Hashable]] = None):
    """Turns an event handler into an action using given name as event's name"""
    # pylint: disable=protected-access

    action._counter = getattr(action, "_counter", 0) + 1

    def _decorator_action(handler: Callable) -> Action:
        @functools.wraps(handler)
        def _wrapper_action(*args, **kwargs):
            handler(*args, **kwargs)

        _wrapper_action.name = (
            handler.__name__ + "_" + str(action._counter) if name is None else name
        )
        _wrapper_action.keys = keys

        return _wrapper_action

    return _decorator_action


class Dispatcher:
    """Dispatcher an event's values to a matching handler."""

    def __init__(self):
        self._handlers: Dict[str, Action] = {}

    def register(self, handler: Action) -> Self:
        """Registers given action's handler by its name and keys."""
        self._handlers.setdefault(handler.name, []).append(handler)
        if handler.keys is not None:
            for key in handler.keys:
                self._handlers.setdefault(key, []).append(handler)
        return self

    def dispatch(self, event, values) -> bool:
        """Returns True if a handler was found and invoked for given event."""
        if isinstance(event, tuple):
            # TODO are there other type of "tuple"-events?
            name = event[0] if "+CLICKED+" == event[1] else None
        else:
            if 2 == len(menu_event := event.rsplit(sg.MENU_KEY_SEPARATOR, 1)):
                _, name = menu_event  # extract the key from a menu-item event having a name
            else:
                name = event

        if (handlers := self._handlers.get(name, None)) is not None:
            for handler in handlers:
                handler(values)
            return True
        return False

    def loop(self, window: sg.Window, timeout_ms=None, timeout_key=sg.TIMEOUT_KEY):
        """Process window's events and values until the Exit event or given timeout"""
        log = logging.getLogger("PSGA")
        while True:
            event, values = window.read(timeout_ms, timeout_key)
            log.debug("event %s, values: %s", event, values)

            if event in {sg.WIN_CLOSED, "Exit"}:
                break

            if self.dispatch(event, values):
                continue

            log.warning("Unhandled event: %s", event)


class Controller:
    """Groups and registers actions to a dispatcher."""

    # pylint: disable=too-few-public-methods

    def __init__(self, dispatcher: Dispatcher, window: Optional[sg.Window] = None):
        self.window = window
        for method in [
            func
            for func in map(lambda name: getattr(self, name), dir(self))
            if callable(func) and hasattr(func, "name")
        ]:
            dispatcher.register(method)
